from zcm import *

class Component_1(Component):
    def __init__(self):
        Component.__init__(self)
        self.register_timer_operation("timer_1_function", self.timer_1_function)
        self.register_subscriber_operation("subscriber_function", self.subscriber_function)

    def timer_1_function(self):
        self.publisher("publisher_port").send("Component_1")
        print "Component_1 : Timer : Published message : Component_1"

    def subscriber_function(self, received_message):
        print "Component_1 : Subscriber : Received message:", received_message



#!/usr/bin/env python
"""component.py: This file contains the Component class."""
from operation_queue import OperationQueue
from timer import Timer
from publisher import Publisher
from subscriber import Subscriber
from client import Client
from server import Server

__author__ = "Pranav Srinivas Kumar"
__copyright__ = "Copyright 2016, Pranav Srinivas Kumar"
__credits__ = ["Pranav Srinivas Kumar"]
__license__ = "MIT"
__version__ = "1.0.0"
__maintainer__ = "Pranav Srinivas Kumar"
__email__ = "pranav.srinivas.kumar@gmail.com"
__status__ = "Production"

# Component class
class Component():
    def __init__(self, operation_queue_size=10000):
        """Construct a component - Prepare the operation queue"""
        self.operation_queue = OperationQueue(operation_queue_size)     
        self.timers = []
        self.publishers = []
        self.subscribers = []
        self.clients = []
        self.servers = []
        self.timer_functions = {}
        self.subscriber_functions = {}
        self.server_functions = {}
        self.executor_thread = None

    def timer(self, timer_name):
        """Return a component timer by name"""
        for timer in self.timers:
            if timer.name == timer_name:
                return timer
        return None

    def publisher(self, publisher_name):
        """Return a component publisher by name"""
        for publisher in self.publishers:
            if publisher.name == publisher_name:
                return publisher
        return None

    def subscriber(self, subscriber_name):
        """Return a component subscriber by name"""
        for subscriber in self.subscribers:
            if subscriber.name == subscriber_name:
                return subscriber
        return None

    def client(self, client_name):
        """Return a component client by name"""
        for client in self.clients:
            if client.name == client_name:
                return client
        return None

    def server(self, server_name):
        """Return a component server by name"""
        for server in self.servers:
            if server.name == server_name:
                return server
        return None

    def add_timer(self, new_timer):
        """Add a new timer to this component"""
        self.timers.append(new_timer)

    def add_publisher(self, new_publisher):
        """Add a new publisher to this component"""
        self.publishers.append(new_publisher)

    def add_subscriber(self, new_subscriber):
        """Add a new subscriber to this component"""
        self.subscribers.append(new_subscriber)

    def add_client(self, new_client):
        """Add a new client to this component"""
        self.clients.append(new_client)

    def add_server(self, new_server):
        """Add a new server to this component"""
        self.servers.append(new_server)

    def configure_publishers(self, publisher_dictionary):
        """Configure all component publishers"""
        for key, value in publisher_dictionary.iteritems():
            this_publisher = self.publisher(key)
            if not (this_publisher == None):
                this_publisher.bind(value)

    def configure_subscribers(self, subscriber_dictionary):
        """Configure all component subscribers"""
        for key, value in subscriber_dictionary.iteritems():
            this_subscriber = self.subscriber(key)
            if not (this_subscriber == None):
                this_subscriber.connect(value)

    def configure_clients(self, client_dictionary):
        """Configure all component clients"""
        for key, value in client_dictionary.iteritems():
            this_client = self.client(key)
            if not (this_client == None):
                this_client.connect(value)

    def configure_servers(self, server_dictionary):
        """Configure all component servers"""
        for key, value in server_dictionary.iteritems():
            this_server = self.server(key)
            if not (this_server == None):
                this_server.bind(value)

    def register_timer_operation(self, operation_name, operation_function):
        """Register a timer operation"""
        self.timer_functions[operation_name] = operation_function

    def register_subscriber_operation(self, operation_name, operation_function):
        """Register a subscriber operation"""
        self.subscriber_functions[operation_name] = operation_function

    def register_server_operation(self, operation_name, operation_function):
        """Register a server operation"""
        self.server_functions[operation_name] = operation_function

    def spawn(self):
        """Spawn the component executor thread and all port-specific threads"""
        self.executor_thread = self.operation_queue.spawn()
        for timer in self.timers:
            timer.start()
        for subscriber in self.subscribers:
            subscriber.start()
        for server in self.servers:
            server.start()
        return self.executor_thread
    



from zcm import *

class Component_2(Component):
    def __init__(self):
        Component.__init__(self)
        self.register_timer_operation("timer_function", self.timer_function)
        self.register_server_operation("server_function", self.server_function)

    def timer_function(self):
        self.publisher("publisher_port").send("Component_2")
        print "Component_2 : Timer : Published Message : Component_2"

    def server_function(self, received_request):
        print "Component_2 : Server : Received message:", received_request
        self.publisher("publisher_port").send("Component_2")
        print "Component_2 : Server : Published message : Component_2"
        return "Component_2_ACK"



from component import Component

class component_1(Component):
    def __init__(self):
        Component.__init__(self)
        self.register_timer_operation("timer_1_function", 
                                      self.timer_1_function)
        self.register_subscriber_operation("subscriber_function",
                                           self.subscriber_function)

    def timer_1_function(self):
        self.publisher("publisher_port").send("Component_1")
        print "Component 1 : Timer : Published message: Component_1"

    def subscriber_function(self, received_message):
        print "Component 1 : Subscriber : Received message:", received_message 


class component_2(Component):
    def __init__(self):
        Component.__init__(self)
        self.register_timer_operation("timer_function",
                                      self.timer_function)
        self.register_server_operation("server_function",
                                       self.server_function)

    def timer_function(self):
        self.publisher("publisher_port").send("Component_2")
        print "Component_2 : Timer : Published message: Component_2"

    def server_function(self, received_request):
        print "Component_2 : Server : Received request:", received_request
        self.publisher("publisher_port").send("Component_2")
        print "Component_2 : Server : Published message: Component_2"
        return "Component_2"

class component_3(Component):
    def __init__(self):
        Component.__init__(self)
        self.register_timer_operation("timer_function", self.timer_function)

    def timer_function(self):
        response = self.client("client_port").call("Component_3")
        print "Component_3 : Timer : Called Component_2::Server : Received : ", response



#!/usr/bin/env python
"""timer.py: This file contains the Timer class."""
import sched, time
from threading import Thread, Lock
from operation_queue import OperationQueue, TimerOperation

__author__ = "Pranav Srinivas Kumar"
__copyright__ = "Copyright 2016, Pranav Srinivas Kumar"
__credits__ = ["Pranav Srinivas Kumar"]
__license__ = "MIT"
__version__ = "1.0.0"
__maintainer__ = "Pranav Srinivas Kumar"
__email__ = "pranav.srinivas.kumar@gmail.com"
__status__ = "Production"

# Timer class - Create a periodic timer and bind the timer to a callback operation
class Timer():
    def __init__(self, name, priority, period, 
                 operation_function, operation_queue):
        """
        Create a timer

        Keyword arguments:
        name - Name of the timer
        priority - Priority of the timer
        period - Period of the timer (in seconds)
        operation_function - Operation to which the timer is bound
        operation_queue - The operation queue object
        """
        self.name = name
        self.period = period
        self.priority = priority
        self.operation_function = operation_function
        self.operation_queue = operation_queue
        self.period_mutex = Lock()
        self.func_mutex = Lock()

    def operation(self):
        """
        Timer thread function

        Behavior:
        (1) Wait for timer expiry
        (2) Create a timer operation
        (3) Enqueue onto operation_queue
        (4) Goto step (1)
        """
        while(True):
            self.period_mutex.acquire()
            start = time.time()
            while((time.time() - start) < self.period):
                pass
            self.period_mutex.release()
                
            self.func_mutex.acquire()
            new_operation = TimerOperation(self.name, self.priority, 
                                           self.operation_function)
            self.operation_queue.enqueue(new_operation)
            self.func_mutex.release()
            
    def change_period(self, new_period):
        """
        Change the timer period

        Keyword arguments:
        new_period - New timer period in seconds
        """
        self.period_mutex.acquire()
        self.period = new_period
        self.period_mutex.release()

    def rebind_operation_function(self, new_operation_function):
        """
        Rebind the timer operation function

        Keyword arguments:
        new_operation_function - New timer function to handle timer expiry
        """        
        self.func_mutex.acquire()
        self.operation_function = new_operation_function
        self.func_mutex.release()

    def spawn(self):
        """Spawn a new thread for the timer."""
        return Thread(target = self.operation)

    def start(self):
        """Start the timer thread."""
        timer_thread = self.spawn()
        timer_thread.start()

        
        



#!/usr/bin/env python
"""server.py: This file contains the Server class."""
import zmq
from threading import Thread, Lock
from operation_queue import OperationQueue, ServerOperation

__author__ = "Pranav Srinivas Kumar"
__copyright__ = "Copyright 2016, Pranav Srinivas Kumar"
__credits__ = ["Pranav Srinivas Kumar"]
__license__ = "MIT"
__version__ = "1.0.0"
__maintainer__ = "Pranav Srinivas Kumar"
__email__ = "pranav.srinivas.kumar@gmail.com"
__status__ = "Production"

# Server class
class Server():
    def __init__(self, name, priority, endpoints, 
                 operation_function, operation_queue):
        """
        Create a server

        Keyword arguments:
        name - Name of the timer
        priority - Priority of the subscriber
        endpoints - A list of endpoint strings
        operation_function - Operation function of the subscriber
        operation_queue - The operation queue object
        """
        self.name = name
        self.priority = priority
        self.endpoints = endpoints
        self.operation_function = operation_function
        self.operation_queue = operation_queue
        self.context = zmq.Context()
        self.server_socket = self.context.socket(zmq.REP)
        for endpoint in self.endpoints:
            self.server_socket.bind(endpoint)
        self.ready = True
        self.func_mutex = Lock()

    def bind(self, new_endpoints):
        """
        Bind to a new set of endpoints

        Keyword arguments:
        new_endpoints - A new set of endpoints to connect to
        """
        self.endpoints = new_endpoints
        self.context = zmq.Context()
        self.server_socket = self.context.socket(zmq.REP)
        for endpoint in self.endpoints:
            self.server_socket.bind(endpoint)
        self.ready = True

    def add_connection(self, new_connection):
        """
        Add a new endpoint
        
        Keyword arguments:
        new_connection - New connection address to connect to
        """
        self.server_socket.bind(new_connection)
    
    def recv(self):
        """
        Thread function of the server

        Behavior:
        (1) Wait for a new request on the server zmq socket
        (2) Create a server operation
        (3) Enqueue onto the operation queue
        (4) Goto step (1)
        """
        while(True):
            while(self.ready == False):
                pass
            received_request = self.server_socket.recv()
            self.ready = False
            if (len(received_request) > 0):
                self.func_mutex.acquire()
                def bind(func, *args, **kwargs):
                    return lambda: func(*args, **kwargs)
                new_operation = ServerOperation(self.name, self.priority, 
                                                bind(self.operation_function, 
                                                     received_request),
                                                self)
                self.operation_queue.enqueue(new_operation)
                self.func_mutex.release()

    def rebind_operation_function(self, new_operation_function):
        """
        Rebind the subscriber operation function
        
        Keyword arguments:
        new_operation_function - New server function to be handled upon recv()
        """
        self.func_mutex.acquire()
        self.operation_function = new_operation_function
        self.func_mutex.release()

    def spawn(self):
        """Spawn a new thread for the server."""        
        return Thread(target = self.recv)

    def start(self):
        """Start the server thread."""
        server_thread = self.spawn()
        server_thread.start()        



#!/usr/bin/env python
"""actor.py: This file contains the Actor class."""
import json
import importlib
from timer import Timer
from publisher import Publisher
from subscriber import Subscriber
from client import Client
from server import Server

__author__ = "Pranav Srinivas Kumar"
__copyright__ = "Copyright 2016, Pranav Srinivas Kumar"
__credits__ = ["Pranav Srinivas Kumar"]
__license__ = "MIT"
__version__ = "1.0.0"
__maintainer__ = "Pranav Srinivas Kumar"
__email__ = "pranav.srinivas.kumar@gmail.com"
__status__ = "Production"

# Actor class
class Actor():
    def __init__(self):
        """Construct an actor - Prepare an empty list of component instances"""
        self.component_instances = []

    def configure(self, configuration_file):
        """
        Configure the component instances list
        
        Keyword arguments:
        configuration_file - JSON configuration file to parse
        """
        with open(configuration_file) as data_file:
            root = json.load(data_file)
            for instance in root["Component Instances"]:
                instance_source = instance["Definition"]
                module_name, class_name = instance_source.split(".") 
                module = importlib.import_module(module_name) 
                component_instance = getattr(module, class_name)() 

                publishers_config_map = {}
                subscribers_config_map = {}
                clients_config_map = {}
                servers_config_map = {}

                if "Timers" in instance.keys():
                    for timer in instance["Timers"]:
                        timer_name = timer["Name"]
                        timer_priority = timer["Priority"]
                        timer_period = timer["Period"]
                        timer_operation = timer["Function"]
                        new_timer = Timer(timer_name,
                                          timer_priority,
                                          timer_period,
                                          component_instance.timer_functions[timer_operation],
                                          component_instance.operation_queue)
                        component_instance.add_timer(new_timer)

                if "Publishers" in instance.keys():
                    for publisher in instance["Publishers"]:
                        publisher_name = publisher["Name"]
                        for endpoint in publisher["Endpoints"]:                        
                            if publisher_name not in publishers_config_map.keys():
                                publishers_config_map[publisher_name] = []
                                publishers_config_map[publisher_name].append(endpoint)
                                new_publisher = Publisher(publisher_name)
                                component_instance.add_publisher(new_publisher)

                if "Subscribers" in instance.keys():
                    for subscriber in instance["Subscribers"]:
                        subscriber_name = subscriber["Name"]
                        subscriber_priority = subscriber["Priority"]
                        subscriber_filter = subscriber["Filter"]
                        subscriber_operation = subscriber["Function"]
                        for endpoint in subscriber["Endpoints"]:
                            if subscriber_name not in subscribers_config_map.keys():
                                subscribers_config_map[subscriber_name] = []
                                subscribers_config_map[subscriber_name].append(endpoint)
                                new_subscriber = Subscriber(subscriber_name,
                                                            subscriber_priority,
                                                            subscriber_filter,
                                                            subscribers_config_map[subscriber_name],
                                                            component_instance.\
                                                            subscriber_functions[subscriber_operation],
                                                            component_instance.operation_queue)
                                component_instance.add_subscriber(new_subscriber)

                if "Clients" in instance.keys():
                    for client in instance["Clients"]:
                        client_name = client["Name"]
                        for endpoint in client["Endpoints"]:
                            if client_name not in clients_config_map.keys():
                                clients_config_map[client_name] = []
                                clients_config_map[client_name].append(endpoint)
                                new_client = Client(client_name)
                                component_instance.add_client(new_client)

                if "Servers" in instance.keys():
                    for server in instance["Servers"]:
                        server_name = server["Name"]
                        server_priority = server["Priority"]
                        server_operation = server["Function"]
                        for endpoint in server["Endpoints"]:
                            if server_name not in servers_config_map.keys():
                                servers_config_map[server_name] = []
                                servers_config_map[server_name].append(endpoint)
                                new_server = Server(server_name,
                                                    server_priority,
                                                    servers_config_map[server_name],
                                                    component_instance.server_functions[server_operation],
                                                    component_instance.operation_queue)
                                component_instance.add_server(new_server)

                component_instance.configure_publishers(publishers_config_map)
                component_instance.configure_subscribers(subscribers_config_map)
                component_instance.configure_clients(clients_config_map)
                component_instance.configure_servers(servers_config_map)
                self.component_instances.append(component_instance)

    def run(self):
        """Spawn all component instances"""
        for instance in self.component_instances:
            instance_thread = instance.spawn()
            instance_thread.start()



#!/usr/bin/python

from operation_queue import TimerOperation, SubscriberOperation, ServerOperation, OperationQueue
from timer import Timer
from publisher import Publisher
from subscriber import Subscriber
from client import Client
from server import Server
from component import Component
from actor import Actor
import argparse
import os, sys, inspect

pwd = os.getcwd()
if pwd not in sys.path:
    sys.path.insert(0, pwd)

def main():
    """Entry point for the application script"""

    parser = argparse.ArgumentParser(description=\
                                     'functionality: Spawn a ZCM Actor', 
                                     formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument('--config', nargs='?', default=None, help='Name of configuration file')
    args = vars(parser.parse_args())

    if not (args['config'] == None):
        my_actor = Actor()
        my_actor.configure(args['config'])
        my_actor.run()



from zcm import *

class ServerComponent(Component):
    def __init__(self):
        """Register the server operation"""
        Component.__init__(self)
        self.register_server_operation("server_function", 
                                       self.server_function)

    def server_function(self, request):
        """Receive requests and respond to remote clients"""
        print "Server : Received message:", request 
        return "ACK"






#!/usr/bin/env python
"""publisher.py: This file contains the Publisher class."""
import zmq

__author__ = "Pranav Srinivas Kumar"
__copyright__ = "Copyright 2016, Pranav Srinivas Kumar"
__credits__ = ["Pranav Srinivas Kumar"]
__license__ = "MIT"
__version__ = "1.0.0"
__maintainer__ = "Pranav Srinivas Kumar"
__email__ = "pranav.srinivas.kumar@gmail.com"
__status__ = "Production"

# Publisher class
class Publisher():
    def __init__(self, name, endpoints = None):
        """
        Create a publisher

        Keyword arguments:
        name - Name of the timer
        endpoints - A list of endpoint strings
        """
        self.name = name
        self.endpoints = None
        self.context = None
        self.publisher_socket = None
        if not (endpoints == None):
            self.endpoints = endpoints
            self.context = zmq.Context()
            self.publisher_socket = self.context.socket(zmq.PUB)
            for endpoint in self.endpoints:
                self.publisher_socket.bind(endpoint)

    def bind(self, new_endpoints):
        """
        Bind the publisher to a new set of endpoints

        Keyword arguments:
        new_endpoints - New set of endpoints as a list
        """
        self.endpoints = new_endpoints
        self.context = zmq.Context()
        self.publisher_socket = self.context.socket(zmq.PUB)
        for endpoint in new_endpoints:
            self.publisher_socket.bind(endpoint)

    def add_connection(self, new_connection):
        """
        Add a new endpoint to the publisher

        Keyword arguments:
        new_connection - New endpoint to bind to
        """
        self.publisher_socket.bind(new_connection)

    def send(self, message):
        """
        Publish a new message
        
        Keyword arguments:
        message - The message string. Serialize complex objects to strings with e.g. protobuf
        """
        self.publisher_socket.send(message)



from zcm import *

class ClientComponent(Component):
    def __init__(self):
        """Register a timer operation"""
        Component.__init__(self)
        self.register_timer_operation("timer_function", 
                                      self.timer_function)

    def timer_function(self):
        """Request the services of a remote server using the client_port"""
        print "Client Timer : Sending message: client_timer_message" 
        response = self.client("client_port").call("client_timer_message")
        print "Client Timer : Received response :", response



from zcm import *

class SubscriberComponent(Component):
    def __init__(self):
        """Register a subscriber operation"""
        Component.__init__(self)
        self.register_subscriber_operation("subscriber_function",
                                           self.subscriber_function)

    def subscriber_function(self, message):
        print "Subscriber : Received Message:", message



from zcm import *

class PublisherComponent(Component):
    def __init__(self):
        """Register a timer operation"""
        Component.__init__(self)
        self.register_timer_operation("timer_1_function",
                                      self.timer_1_function)

    def timer_1_function(self):
        self.publisher("timer_pub_1").send("timer_1_message")
        print "Publisher : Published message: timer_1_message"




from zcm import *

class TimersComponent(Component):
    def __init__(self):
        """Register all timer operations"""
        Component.__init__(self)
        self.register_timer_operation("timer_1_function", self.timer_1_function)
        self.register_timer_operation("timer_2_function", self.timer_2_function)
        self.register_timer_operation("timer_3_function", self.timer_3_function)
        self.register_timer_operation("timer_4_function", self.timer_4_function)
        self.register_timer_operation("timer_5_function", self.timer_5_function)

    def timer_1_function(self):
        print "Timer 1 expiry handled" 

    def timer_2_function(self):
        print "Timer 2 expiry handled" 

    def timer_3_function(self):
        print "Timer 3 expiry handled" 

    def timer_4_function(self):
        print "Timer 4 expiry handled" 

    def timer_5_function(self):
        print "Timer 5 expiry handled" 



#!/usr/bin/env python
"""operation_queue.py: This file contains classes for supported operation types & the component operation queue.
"""
from Queue import PriorityQueue
from threading import Thread, Lock
import zmq

__author__ = "Pranav Srinivas Kumar"
__copyright__ = "Copyright 2016, Pranav Srinivas Kumar"
__credits__ = ["Pranav Srinivas Kumar"]
__license__ = "MIT"
__version__ = "1.0.0"
__maintainer__ = "Pranav Srinivas Kumar"
__email__ = "pranav.srinivas.kumar@gmail.com"
__status__ = "Production"

# Timer Operation Class - This operation is created when ever a timer expires
class TimerOperation():
    def __init__(self, name, priority, operation_function):
        """
        Construct a timer operation
        
        Keyword arguments:
        name -- Name of the operation
        priority -- Priority of the operation
        operation_function -- Timer function
        """
        self.name = name
        self.priority = priority
        self.operation_function = operation_function

    def execute(self):
        """Execute the timer operation function i.e. the callback."""
        self.operation_function()

# Subscriber Operation class - This operation is created when a subscriber receives messages
class SubscriberOperation():
    def __init__(self, name, priority, operation_function):
        """
        Construct a subscriber operation
        
        Keyword arguments:
        name -- Name of the operation
        priority -- Priority of the operation
        operation_function -- Timer function
        """
        self.name = name
        self.priority = priority
        self.operation_function = operation_function

    def execute(self):
        """Execute the subscriber operation function i.e. the callback."""
        self.operation_function()

# Server Operation class - This operation is created when a server receives requests
class ServerOperation():
    def __init__(self, name, priority, operation_function, server):
        """
        Construct a subscriber operation
        
        Keyword arguments:
        name -- Name of the operation
        priority -- Priority of the operation
        operation_function -- Timer function
        server -- The server object used to reply to the client when the operation completes
        """
        self.name = name
        self.priority = priority
        self.operation_function = operation_function
        self.server = server

    def execute(self):
        """Execute the server operation function and respond back to the client."""
        response = self.operation_function()
        if not (self.server == None):
            if not (self.server.server_socket == None):
                self.server.server_socket.send(response)
        self.server.ready = True

# Operation Queue class - The component operation (priority) queue that holds operations
class OperationQueue():
    def __init__(self, queue_size = 1000):
        """
        Construct the priority queue and prepare a queue mutex

        Keyword arguments:
        queue_size - Size of the operation queue i.e. maximum number of operations
        """
        self.operation_queue = PriorityQueue(queue_size)
        self.queue_mutex = Lock()

    def enqueue(self, new_operation):
        """
        Enqueue a new operation onto the operation queue

        Keyword arguments:
        new_operation - New operation to be enqueued.
        """
        self.queue_mutex.acquire()
        self.operation_queue.put((new_operation.priority, new_operation))
        self.queue_mutex.release()

    def dequeue(self):
        """Dequeue the front operation from the operation queue."""
        return self.operation_queue.get()

    def empty(self):
        """Return true if the operation queue is empty; else return false."""
        return self.operation_queue.empty()

    def process(self):
        """Process the operation queue - Execute operations in the queue if queue not empty."""
        while(True):
            if (not self.operation_queue.empty()):
                self.queue_mutex.acquire()
                first_operation = self.dequeue()
                self.queue_mutex.release()
                first_operation[1].execute()

    def spawn(self):
        """Spawn the component executor thread that processes the queue."""
        return Thread(target = self.process)



from zcm import *

class Component_3(Component):
    def __init__(self):
        Component.__init__(self)
        self.register_timer_operation("timer_function", self.timer_function)
        
    def timer_function(self):
        response = self.client("client_port").call("Component_3")
        print "Component_3 : Timer : Called Component_2::Server : Received:", response
